/******************** torture.c *********************
*  Torture testing for compilers following the ANSI *
*  standard. This programs reads spec records and   *
*  generates code for compilation and running.      *
*  (c) Copyright 1988 C Gazette by A. Binstock      *
****************************************************/
/* Validated for Ecosoft, MSC, Turbo, Watcom */

/* Datalite and MetaWare can't handle __DATE__
   Lattice gets lost, Desmet compiles but won't run  */

#include <stdio.h>
#include <ctype.h>

#define TAB 8               /* Tab width in columns */

FILE *fopen(), *specfile, *codefile;

     /* Various parameters which the user can specify */
int  args,    cases,    defines,  define_args,  enums,
     externs, id_len,   locals;

char id_name [196];         /* 195 + 1 byte type + 4 byte sequence = */
                            /* 200 byte max. identifier name length  */

void do_switches (char *), write_rest (void), write_prologue (char *[]),
     write_body (void), do_ident_name (void);

main (int argc, char * argv[])
{
    char spec_line[82];

    if (argc < 2)
    {
        puts ("This program will generate a test file called: TESTCODE.C");
        puts ("Usage: torture filename.ext");
        puts ("  where filename contains the specs for the code generation");
        exit (2);
    }
    else
    {
        if ((specfile = fopen (argv[1], "r")) == NULL)
        {
            printf ("Can't open spec file: %s\n", argv[1]);
            exit (3);
        }
        else
        {
            if ((codefile = fopen ("testcode.c", "w")) == NULL)
            {
                puts ("Can't open output code file TESTCODE.C");
                exit (4);
            }
        }
    }

    while (fgets (spec_line, 81, specfile) != NULL)
    {
        do_switches (spec_line);
    }

    do_ident_name ();
    write_prologue(argv);
    write_body();
    write_rest ();
    fflush (codefile);
    fclose (codefile);              /* Unfortunately, fcloseall() from    */
    fclose (specfile);              /* UNIX is not part of ANSI library.  */
    return (10);
}

void do_switches (char *line)       /* Read each line of the specs file:  */
{                                   /* semi-colon marks ASM-type comment. */
    char setting [80];              /* switches are in the form x9999     */
    int  i = 0;                     /* where x is the statement type and  */
    char c;                         /* and 9999 is the number of occurs.  */
    c = *line;                      /* Some switches are: xx9999          */
    while (c != '\0' && c != ';' && c != '\r')  /* ; = comment until EOL  */
    {
         setting [i++] = c;
         c = *(++line);
         if (i == 80)
         {
            setting [--i] = '\0';
            break;
         }
    }
    setting[i] = '\0';

    line = setting;
    switch (tolower (*line))
    {
        case 'a':                        /* Arguments passed to function  */
            args    = atoi (++line);
            break;
        case 'c':                        /* Case statements in a switch() */
            cases   = atoi (++line);
            break;
        case 'd':
        {
            if (isdigit (*++line))       /* How many #define statements   */
               defines = atoi (line);
            else
            if (tolower(*line) == 'a')   /* How many args in one #define  */
               define_args = atoi (++line);
        }
            break;
        case 'e':
        {
            char next;
            ++line;
            next = tolower(*line);

            if (next == 'x')             /* How many externs              */
                externs = atoi (++line);
            else
            if (next == 'n')             /* How many enums                */
                enums   = atoi (++line);
        }
            break;
        case 'i':                        /* Length of identifiers         */
            id_len  = atoi (++line);     /* Must be at least 5 chars      */
            id_len  = id_len > 5 ? id_len - 5 : 0;
            break;
    }
}

/* This next function writes the body if the code. It starts by doing
    externs, #defines, enums, and all items which can be done before main().
    main() is defined and then the body of main() is tested for and
    created in logical order.                                             */

void write_body(void)
{
    while (externs)          /*  First generate the externs               */
    {
        fprintf (codefile, "extern int %se%04d;\n", id_name, externs);
        --externs;
    }
    fprintf (codefile, "\n");

    while (defines)          /*  Then generate the defines                */
    {
        fprintf (codefile, "#define %sd%04d  1\n", id_name, defines);
        --defines;
    }

    if (enums)               /* The do the enums                          */
    {
        int spacing;
        spacing = ((80 - TAB) / (id_len + 5 + 2));
        if (spacing < 1)
            spacing = 1;

        fprintf (codefile, "\nenum test_enum {\n\t");

        while (enums > 1)
        {
            fprintf (codefile, "%sz%04d, ", id_name, enums);
            enums--;
            if (enums % spacing == 0)
                fputs ("\n\t", codefile);
        }
        fprintf (codefile, "\n\t%sz%04d ", id_name, enums);
        fprintf (codefile, "\n}; /* End of enums */\n");
    }

    fputs ("\nmain (int argc, char * argv[])\n{\n", codefile);    /* main() */
    fputs ("\tint main_i = 1;\n", codefile);

    if (args)
    {                                /* If more locals are already defined, */
                                     /* use them as args, else create args. */
        int args_to_gen = (args > locals ? (args - locals) : 0);
        if (args_to_gen)
        {
            fputs ("\n\tlong int", codefile);
            while (args_to_gen > 1)
            {
                fprintf (codefile, "\n\t\t%sa%04d,", id_name, args_to_gen);
                args_to_gen--;
            }
            fprintf (codefile, "\n\t\t%sa%04d;", id_name, args_to_gen);
        }
    }

    if (cases)
    {
        fputs ("\n\tswitch (argc)\n\t{\n", codefile);
        while (cases)
        {
            fprintf (codefile, "\t\tcase %d:  main_i--; break;\n", cases);
            cases--;
        }
        fputs ("\t}\n", codefile);
    }

    if (args)
    {                     /* If more locals are already defined, */
                          /* use them as args, else create args. */
        int spacing;
        int args_to_gen = (args > locals ? (args - locals) : 0);

        spacing = (80 - (2 * TAB)) / (id_len + 5);
        if (spacing < 1)
            spacing = 1;

        if (args_to_gen)
        {
            fputs ("arg_func (", codefile);
            while (args_to_gen > 1)
            {
                fprintf (codefile, "%sa%04d, ", id_name, args_to_gen);
                args_to_gen--;
                if ((args_to_gen % spacing) == 0)
                    fputs ("\n\t\t", codefile);
            }
            fprintf (codefile, "\n\t\t%sa%04d);\n", id_name, args_to_gen);
        }
    }
    fputs ("}\n\n", codefile);
}


void write_prologue(char *argv[])           /* Writes the file header */
{
    fputs ("/* Torture Test of Compiler. Version 1.2.3    */\n", codefile);
    fprintf (codefile, "/* %s.   Generated from %s */\n", __DATE__, argv[1]);
    fputs ("/* by code generator from The C Gazette staff */\n\n", codefile);
    fputs ("\n#include <stdio.h>\n\n", codefile);
}

void write_rest (void)
{
    if (args)
    {

        int args_to_gen, spacing;

        spacing = (80 - (2 * TAB)) / (id_len + 5);
        if (spacing < 1)
            spacing = 1;

        fputs
            ("/* Function accepts many arguments and just prints message */\n",
             codefile);
        fputs ("arg_func (", codefile);
        args_to_gen = (args > locals ? (args - locals) : 0);
        if (args_to_gen)
        {
            while (args_to_gen > 1)
            {
                fprintf (codefile, "long %sa%04d, ", id_name, args_to_gen);
                args_to_gen--;
                if ((args_to_gen % spacing) == 0)
                    fputs ("\n\t\t", codefile);
            }
            fprintf (codefile, "\n\t\tlong %sa%04d)\n\n{", id_name, args_to_gen);
            fputs ("\t\tputs (\"Made it to called arg function\");\n}\n\n",
                    codefile);
        }
    }
}

void do_ident_name(void)   /* Creates identifier prefix of length id_len */
{
    if (!id_len)
        strcpy (id_name, " ");  /* Initialize id_name to 1 space minimum */
    else
    {                           /* This code relies on ANSI draft conca- */
        strcpy (id_name,        /* nation of adjacent strings.           */
        "aa005bb010cc015dd020ee025ff030gg035hh040ii045jj050kk055ll060mm065"
        "nn070oo075pp080qq085rr090ss095tt100uu105vv110ww115xx120yy125zz130"
        "aa135bb140cc145dd150ee155ff160gg165hh170ii175jj180kk185ll190mm195");
        id_name[id_len] = '\0';
    }
}
